@@ -10,22 +10,24 @@ import pandas as pd
 import pyqtgraph as pg
 from matplotlib import cm
 from matplotlib.pyplot import colormaps
+from pyqtgraph import GraphicsScene
 from pyqtgraph.Qt import QtWidgets, QtGui, QtCore
-# Taken directly from https://stackoverflow.com/a/20610786/9463643
 from pyqtgraph.graphicsItems.LegendItem import ItemSample
 from pyqtgraph.graphicsItems.ScatterPlotItem import drawSymbol
 from pyqtgraph.graphicsItems.ViewBox.ViewBoxMenu import ViewBoxMenu
 from pyqtgraph.parametertree import Parameter
 from typing_extensions import Literal
+from numpy.distutils.misc_util import is_sequence
+
 from utilitys.fns import nameFormatter, pascalCaseToTitle, params_flattened, \
   dynamicDocstring, getParamChild
 from utilitys.misc import CompositionMixin
 from utilitys.typeoverloads import FilePath
-
 from . import params
 from .constants import PrjEnums
 from .params.prjparam import PrjParam
 
+# Taken directly from https://stackoverflow.com/a/20610786/9463643
 try:
   from pyqtgraph.Qt import QtWidgets
   from qtconsole.rich_jupyter_widget import RichJupyterWidget
@@ -459,14 +461,14 @@ class HasEasyChild(QtWidgets.QMainWindow):
 
 
 class ImageViewer(CompositionMixin, pg.PlotWidget):
-
+  sigMouseMoved = QtCore.Signal(object) # ndarray(int, int) xy pos rel. to image
 
   def __init__(self, imgSrc: np.ndarray=None, **kwargs):
     super().__init__(**kwargs)
     self.pxColorLbl, self.mouseCoordsLbl = None, None
     """
     Set these to QLabels to have up-to-date information about the image coordinates
-    under the mouse (if self.mouseEnabled = True)    
+    under the mouse    
     """
     self.toolsEditor = params.ParamEditor(name='Region Tools')
     vb = self.getViewBox()
@@ -479,7 +481,6 @@ class ImageViewer(CompositionMixin, pg.PlotWidget):
 
     self.setAspectLocked(True)
     vb.invertY()
-    self.setMouseEnabled(True)
 
     # -----
     # IMAGE
@@ -489,18 +490,21 @@ class ImageViewer(CompositionMixin, pg.PlotWidget):
     self.addItem(self.imgItem)
     if imgSrc is not None:
       self.setImage(imgSrc)
-    self.sigSceneMouseMoved.connect(self._onMouseMove)
-    self.mouseEnabled = True
 
-  def _onMouseMove(self, pos: QtCore.QPointF):
+  def mouseMoveEvent(self, ev: QtGui.QMouseEvent):
+    super().mouseMoveEvent(ev)
+    pos = ev.pos()
     relpos = self.imgItem.mapFromScene(pos)
     xyCoord = np.array([relpos.x(), relpos.y()], dtype=int)
     if (self.imgItem.image is None
         or np.any(xyCoord < 0)
-        or np.any(xyCoord > self.imgItem.image.shape[:2][::-1])):
+        or np.any(xyCoord > np.array(self.imgItem.image.shape[:2][::-1]) - 1)):
       return
     cursorClr = self.imgItem.image[xyCoord[1], xyCoord[0],...].astype(int)
+    if cursorClr.ndim == 0:
+      cursorClr = np.array([cursorClr])
     self.updateCursorInfo(xyCoord, cursorClr)
+    self.sigMouseMoved.emit(xyCoord)
 
   def updateCursorInfo(self, xyPos: np.ndarray, pxColor: np.ndarray):
     if pxColor is None: return
@@ -546,13 +550,15 @@ class ImageViewer(CompositionMixin, pg.PlotWidget):
     else:
       self.imgItem.setImage(imgSrc)
 
-  def widgetContainer(self, asMainWin=False, **kwargs):
+  def widgetContainer(self, asMainWin=False, showTools=True, **kwargs):
     """
     Though this is a PlotWidget class, it has a lot of widget children (toolsEditor group, buttons) that are
     not visible when spawning the widget. This is a convenience method that creates a new, outer widget
     from all teh graphical elements of an EditableImage.
 
     :param asMainWin: Whether to return a QMainWindow or QWidget
+    :param showTools: If `showMainWin` is True, this determines whether to show the tools
+      editor with the window
     :param kwargs: Passed to either EasyWidget.buildMainWin or EasyWidget.BuildWidget,
       depending on the value of `asMainWin`
     """
@@ -563,6 +569,8 @@ class ImageViewer(CompositionMixin, pg.PlotWidget):
       wid.statusBar().addWidget(self.mouseCoordsLbl)
       wid.statusBar().addWidget(self.pxColorLbl)
       wid.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.toolsEditor)
+      if showTools:
+        self.toolsEditor.show()
     else:
       wid = EasyWidget.buildWidget(self._widgetContainerChildren(), **kwargs)
     return wid
@@ -600,11 +608,12 @@ class MaskCompositor(ImageViewer):
     # -----
     # Create properties
     # -----
-    self.legend = CompositorLegend(offset=(5,5), horSpacing=5, brush='cccd', pen='k')
+    self.legend = CompositorLegend(offset=(5,5), horSpacing=5, brush='ccce', pen='k')
+
     self.recordDf = pd.DataFrame(columns=['name', 'item', 'scatter', 'opacity']).set_index('name')
 
     self.legendFontArgs = {'size': '11pt', 'color': 'k', 'bold': True}
-    self.nextZ = 2
+    self.curZ = 2
 
     # -----
     # Configure relationships
@@ -612,7 +621,7 @@ class MaskCompositor(ImageViewer):
     self.legend.setParentItem(self.plotItem)
     self.viewbox: pg.ViewBox = self.getViewBox()
 
-    self.masksParam = getParamChild(self.toolsEditor.params, 'Masks')
+    self.masksParam = getParamChild(self.toolsEditor.params, 'Overlays')
     self.allVisible = True
     # for ax in 'left', 'bottom', 'right', 'top':
     #   self.mainImg.plotItem.hideAxis(ax)
@@ -629,7 +638,7 @@ class MaskCompositor(ImageViewer):
                                     name='Toggle Image Visible')
       self.toolsEditor.registerFunc(lambda: self.legend.setVisible(not self.legend.isVisible()),
                                     name='Toggle Legend Visible')
-      self.toolsEditor.registerFunc(self.toggleAllVisible, namePath=('Masks',))
+      self.toolsEditor.registerFunc(self.toggleAllVisible, namePath=('Overlays',))
     if img is not None:
       self.setImage(img)
 
@@ -651,14 +660,16 @@ class MaskCompositor(ImageViewer):
     if clearOverlays:
       self.clearOverlays()
 
+  def setImage(self, *args, **kwargs):
+    super().setImage(*args, **kwargs)
+    self._updateLegendPos()
+
 
   def addMask(self, mask: np.ndarray, name:str=None, ignoreIfBlank=False, update=True):
     if ignoreIfBlank and not np.any(mask):
       return
     newRecord = {}
 
-    if name is None:
-      name = '[No Name]'
     name = self._getUniqueName(name)
 
     curScat = pg.ScatterPlotItem(symbol='s', width=5)
@@ -666,21 +677,37 @@ class MaskCompositor(ImageViewer):
     newRecord['scatter'] = curScat
 
     toAdd = pg.ImageItem(mask)
-    toAdd.setZValue(self.nextZ+1)
+    toAdd.setZValue(self.curZ + 1)
     self.viewbox.addItem(toAdd)
     newRecord['item'] = toAdd
 
     newRecord['opacity'] = -1
 
     rec = pd.Series(newRecord, name=name)
+    self._addRecord(rec, update)
+
+  def _addRecord(self, rec: pd.Series, update=True):
     self.recordDf = self.recordDf.append(rec)
-    self.nextZ += 1
+    self.curZ += 1
 
     self._addItemCtrls(rec)
 
     if update:
       self._updateGraphics()
 
+  def addImageItem(self, item: pg.ImageItem, **kwargs):
+    kwargs.setdefault('scatter', None)
+    kwargs.setdefault('opacity', -1)
+    name = self._getUniqueName(kwargs.pop('name', None))
+    update = kwargs.pop('update', True)
+    newRecord = dict(item=item, **kwargs)
+
+    item.setZValue(self.curZ + 1)
+    self.viewbox.addItem(item)
+
+    rec = pd.Series(newRecord, name=name)
+    self._addRecord(rec, update)
+
   def addMasks(self, masks: t.List[np.ndarray], names: t.Sequence[t.Optional[str]]=None, ignoreIfBlank=False):
     if names is None:
       names = [None]*len(masks)
@@ -688,7 +715,9 @@ class MaskCompositor(ImageViewer):
       self.addMask(mask, name, ignoreIfBlank, update=False)
     self._updateGraphics()
 
-  def _getUniqueName(self, baseName:str):
+  def _getUniqueName(self, baseName:str=None):
+    if baseName is None:
+      baseName = '[No Name]'
     ii = 2
     name = baseName
     while name in self.recordDf.index:
@@ -696,13 +725,14 @@ class MaskCompositor(ImageViewer):
       ii += 1
     return name
 
-  def addLabelImg(self, labelImg: np.ndarray, name='Label', ignoreBackground=True):
+  def addLabelImg(self, labelImg: np.ndarray, names='Label', ignoreBackground=True):
     """
     Splits a label mask into its constituent labels and adds a mask for each unique label.
 
     :param labelImg: Grayscale label mask of integers
-    :param name: Base name of each label. E.g. if 3 labels are present in the image, the
-      resulting masks will be named <name> 1, <name> 2, <name 3>
+    :param names: Base name of each label. E.g. if 3 labels are present in the image, the
+      resulting masks will be named <name> 1, <name> 2, <name 3>. Alternatively,
+      a list of names can be provided
     :param ignoreBackground: Whether a mask should be added for 0-valued labels (typically
       0 is used to denote background / non-label values)
     """
@@ -710,8 +740,9 @@ class MaskCompositor(ImageViewer):
     if ignoreBackground:
       labels = labels[labels > 0]
     newMasks = [labelImg == label for label in labels]
-    newNames = [name]*len(newMasks)
-    self.addMasks(newMasks, newNames)
+    if not is_sequence(names):
+      names = [f'{names} {ii}' for ii in labels]
+    self.addMasks(newMasks, names)
 
 
   def setLegendFontStyle(self, startItemIdx=0, **lblTxtArgs):
@@ -731,23 +762,37 @@ class MaskCompositor(ImageViewer):
       limits: [0,1]
       step: 0.1
     """
-    names = self.recordDf.index
+    maskNames = np.array([s is not None for s in self.recordDf['scatter']])
+    names = self.recordDf.index[maskNames]
     numMasks = len(names)
     cmap = cm.get_cmap(colormap, numMasks)
     colors = cmap(np.arange(numMasks))
-    opacity = np.where(self.recordDf['opacity'] < 0, opacity, self.recordDf['opacity'])
-    colors[:,-1] = opacity
     colors = (colors*255).astype('uint8')
     backgroundClr = [0,0,0,0]
+    # Filter out names without associated scatters
     for name, color in zip(names, colors):
       # Make sure alpha is not carried over
-      brush, pen = pg.mkBrush(color[:-1]), pg.mkPen(color[:-1])
+      brush, pen = pg.mkBrush(color), pg.mkPen(color)
       self.recordDf.at[name, 'scatter'].setData(symbol='s', brush=brush, pen=pen, width=5)
-    for item, color in zip(self.recordDf['item'], colors):
+    for name, color in zip(names, colors):
       curLut = np.array([backgroundClr, color])
-      item.setOpts(levels=[0,1], lut=curLut)
+      itemOpacity = self.recordDf.at[name, 'opacity']
+      if itemOpacity < 0:
+        itemOpacity = opacity
+      self.recordDf.at[name, 'item'].setOpts(levels=[0,1], lut=curLut, opacity=itemOpacity)
     self.setLegendFontStyle(**self.legendFontArgs)
 
+    # Handle all non-mask items now
+    for name, remainingRec in self.recordDf[~maskNames].iterrows():
+      curOpacity = remainingRec['opacity']
+      if curOpacity < 0:
+        curOpacity = opacity
+      remainingRec['item'].setOpacity(curOpacity)
+
+  def _updateLegendPos(self):
+    imPos = self.imgItem.mapToScene(self.imgItem.pos())
+    self.legend.autoAnchor(imPos)
+
   def _updateGraphics(self):
     self.setOverlayProperties()
 
@@ -756,7 +801,7 @@ class MaskCompositor(ImageViewer):
       self.viewbox.removeItem(imgItem)
     self.recordDf = self.recordDf.drop(self.recordDf.index)
     self.legend.clear()
-    self.toolsEditor.params.child('Masks').clearChildren()
+    self.toolsEditor.params.child('Overlays').clearChildren()
 
   def toggleAllVisible(self):
     for param in self.masksParam:
@@ -764,7 +809,7 @@ class MaskCompositor(ImageViewer):
         param.setValue(not self.allVisible)
     self.allVisible = not self.allVisible
 
-  def save(self, saveFile: FilePath='', cropToViewbox=False, toClipboard=False) -> bool:
+  def save(self, saveFile: FilePath='', cropToViewbox=False, toClipboard=False, floatLegend=False) -> bool:
     """
     :param saveFile:
       helpText: Save destination. If blank, no file is created.
@@ -773,31 +818,64 @@ class MaskCompositor(ImageViewer):
     :param toClipboard: Whether to copy to clipboard
     :param cropToViewbox: Whether to only save the visible portion of the image
     """
-    pm = self.imgItem.getPixmap()
-    painter = QtGui.QPainter(pm)
-    # TODO: Get legend exporting too
+    if self.imgItem.isVisible():
+      saveImg = self.imgItem.getPixmap()
+    else:
+      saveImg = QtGui.QPixmap(*self.imgItem.image.shape[:2][::-1])
+    painter = QtGui.QPainter(saveImg)
+
     visibleMasks = [p.name() for p in params_flattened(self.masksParam) if p.value()]
-    for ii, (name, entry) in enumerate(self.recordDf.iterrows()):
+    for name, item in self.recordDf['item'].iteritems():
       if name in visibleMasks:
-        entry['item'].paint(painter)
-        legItems = self.legend.items[ii]
-        # legItems[0].paint(painter)
-        # legItems[1].paint(painter, None)
-    painter.end()
-    saveImg = pm.toImage()
+        painter.setOpacity(item.opacity())
+        item.paint(painter)
+    painter.setOpacity(1.0)
+
     if cropToViewbox:
       maxBnd = np.array(self.image.shape[:2][::-1]).reshape(-1,1) - 1
       vRange = np.array(self.viewbox.viewRange()).astype(int)
       vRange = np.clip(vRange, 0, maxBnd)
       # Convert range to topleft->bottomright
       pts = QtCore.QPoint(*vRange[:,0]), QtCore.QPoint(*vRange[:,1])
-      saveImg = saveImg.copy(QtCore.QRect(*pts))
+      # Qt doesn't do so well overwriting the original reference
+      origRef = saveImg
+      saveImg = origRef.copy(QtCore.QRect(*pts))
+      painter.end()
+      painter = QtGui.QPainter(saveImg)
+    if self.legend.isVisible():
+      # Wait to paint legend until here in case cropping is active
+      self._paintLegend(painter, floatLegend)
+
     if toClipboard:
-      QtWidgets.QApplication.clipboard().setImage(saveImg)
+      QtWidgets.QApplication.clipboard().setImage(saveImg.toImage())
+
     saveFile = str(saveFile)
     if saveFile and not saveImg.save(str(saveFile)):
       warn('Image compositor save failed', UserWarning)
-    return pm
+    return saveImg
+
+
+  def _paintLegend(self, painter: QtGui.QPainter, floatLegend=False):
+    oldPos = self.legend.pos()
+    try:
+      exportScene = GraphicsScene(parent=None)
+      oldScale = max(np.clip(0.5/np.array(self.imgItem.pixelSize()), 1, np.inf))
+      exportScene.addItem(self.legend)
+      self.legend.setScale(oldScale)
+      # Legend does not scale itself, handle this
+      painter.save()
+      if floatLegend:
+        viewPos = self.viewbox.viewRect()
+        viewX = max(viewPos.x(), 0)
+        viewY = max(viewPos.y(), 0)
+        imPos = self.legend.mapRectToItem(self.imgItem, self.legend.rect())
+        painter.translate(int(imPos.x() - viewX), int(imPos.y() - viewY))
+      exportScene.render(painter)
+      self.legend.setScale(1/oldScale)
+      painter.restore()
+    finally:
+      self.legend.setParentItem(self.plotItem)
+      self.legend.autoAnchor(oldPos, relative=False)
 
   def show_exec(self):
     super().show()
